
import datetime
import enum
import typing as t
from dataclasses import dataclass


@dataclass
class Series:
  " Represents the series that a release belongs to. "

  id: int
  title: str


@dataclass
class Subber:
  " Represents information on the subber that produced a release. "

  id: int
  name: str


@dataclass
class VideoInfo:
  extension: str
  filename: str
  quality: str
  profiles: t.List[str]
  date: datetime.datetime
  comment: str
  size: t.Optional[str]


@dataclass
class Release:
  " Represents a release. "

  id: int
  subber: t.Optional[Subber]
  series: Series
  episode: t.Optional[int]
  version: t.Optional[int]
  video: VideoInfo
  torrent_link: str




__author__ = 'Fading <fading@miracle.org>'
__version__ = '0.1.0'

import json
import os
import posixpath
import requests
import sys
import typing as t
from urllib.parse import urljoin, unquote
from . import model, parser


class ShanaProject:

  DEFAULT_USER_AGENT = f'Python/{sys.version.split()[0]} shanaproject-api ({__version__})'
  LOGIN_URL = 'https://www.shanaproject.com/login/'
  FOLLOWS_URL = 'https://www.shanaproject.com/follows/'
  SEARCH_URL = 'https://www.shanaproject.com/search/'
  IGNORE_URL = 'https://www.shanaproject.com/ajax/ignore/'
  UNIGNORE_URL = 'https://www.shanaproject.com/ajax/queue_rss/'
  DOWNLOAD_URL = 'https://www.shanaproject.com/download/'

  def __init__(self):
    self.session = requests.Session()
    self.session.headers['User-Agent'] = self.DEFAULT_USER_AGENT

  def _ajax_request(self, method: str, url: str, **kwargs: t.Any) -> requests.Response:
    headers = dict(kwargs.pop('headers', None) or {})
    headers['X-Requested-With'] = 'XMLHttpRequest'
    for cookie in self.session.cookies:
      if cookie.name.lower() == 'csrftoken':
        headers['X-CSRFToken'] = cookie.value
        break
    else:
      raise RuntimeError('no csrftoken cookie found, is this a logged in session?')
    return self.session.request(method, url, headers=headers, **kwargs)

  def _get_session_filename(self, id: str) -> str:
    filename = '~/.cache/shanaproject-api-session.{id}.json'
    return os.path.expanduser(filename).format(id=id)

  def save_session(self, id: str) -> None:
    filename = self._get_session_filename(id)
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    with open(filename, 'w') as fp:
      json.dump(dict(self.session.cookies), fp)

  def load_session(self, id: str) -> bool:
    filename = self._get_session_filename(id)
    if os.path.isfile(filename):
      with open(filename) as fp:
        self.session.cookies.update(json.load(fp))
      return True
    return False

  def login(self, username: str, password: str) -> None:
    html = self.session.get(self.LOGIN_URL).text
    csrf_token = parser.get_csrf_middleware_token(html)
    form = {'csrfmiddlewaretoken': csrf_token, 'next': '', 'username': username, 'password': password}
    response = self.session.post(self.LOGIN_URL, data=form)
    response.raise_for_status()

  def follows(self, all_: bool = False, max_pages: int = -1) -> t.Iterable[model.Release]:
    page_num = 1
    page_url = self.FOLLOWS_URL
    if all_:
      page_url += '/all'
    while (max_pages < 0 or page_num <= max_pages) and page_url:
      html = self.session.get(page_url).text
      releases, page_url = parser.parse_search_results(html, self.FOLLOWS_URL)
      yield from releases

  def ignore_release(self, release_id: int) -> None:
    response = self._ajax_request('POST', self.IGNORE_URL, data={'id': release_id})
    response.raise_for_status()

  def unignore_release(self, release_id: int) -> None:
    response = self._ajax_request('POST', self.UNIGNORE_URL, data={'id': release_id})
    response.raise_for_status()

  def download_release(self, release_id: int) -> requests.Response:
    response = self.session.get(self.DOWNLOAD_URL + str(release_id))
    response.raise_for_status()
    return response

  def download_release_to(self, directory: str, release_id: int) -> str:
    " Downloads a release to the *directory* and returns the torrent filename. "

    directory = os.path.expanduser(directory)
    response = self.download_release(release_id)
    filename = posixpath.basename(unquote(response.request.url))
    filename = os.path.join(directory, filename)
    with open(filename, 'wb') as fp:
      fp.write(response.content)
    return filename

  def search(self, title: str = '', subber: str = '', max_pages: int = -1) -> t.Iterable[model.Release]:
    params = {'title': title, 'subber': subber, 'sort': 'date', 'dir': 'Descending'}
    html = self.session.get(self.SEARCH_URL, params=params).text
    page_num = 1
    while (max_pages < 0 or page_num <= max_pages):
      releases, next_page_url = parser.parse_search_results(html, self.SEARCH_URL)
      yield from releases
      if not next_page_url:
        break
      next_page_url = urljoin(self.SEARCH_URL, next_page_url)
      html = self.session.get(next_page_url).text
      page_num += 1




import bs4
import datetime
import re
import typing as t
from urllib.parse import urljoin
from . import model


def get_csrf_middleware_token(html: str) -> str:
  " Find the CSRF token from the login form. "

  soup = bs4.BeautifulSoup(html, 'html.parser')
  input_el = soup.find('input', {'type': 'hidden', 'name': 'csrfmiddlewaretoken'})
  return input_el['value']


def parse_search_results(html: str, base_url: str) -> t.Tuple[t.List[model.Release], t.Optional[str]]:
  """
  Parses the HTML page of the ShanaProject search results page. Returns the parsed releases as
  well as the URL to the next page, if present.
  """

  soup = bs4.BeautifulSoup(html, 'html.parser')

  releases = []
  for i, release_block in enumerate(soup.find_all('div', {'class': 'release_block'})):
    if i == 0: continue
    release = _parse_release_block(release_block)
    release.torrent_link = urljoin(base_url, release.torrent_link)
    releases.append(release)

  list_next = soup.find('div', {'class': 'list_next'}).find('a')
  next_page_url = list_next['href'] if list_next else None

  return releases, next_page_url


def _parse_date_string(ds: str) -> datetime.datetime:
  " Parse a date string like `Aug. 11, 2019, 2:08 a.m.`. "

  match = re.match(r'(\w+)\.?\s+(\d+),\s+(\d+),\s+(\d+):(\d+)\s+(a\.m\.|p\.m\.)', ds)
  if not match:
    raise ValueError(f'unable to parse date string: {ds!r}')

  abbreviated_months = [
    'Jan', 'Feb', 'March', 'April', 'May', 'June', 'July', 'Aug', 'Sept', 'Oct', 'Nov', 'Dec']

  month = abbreviated_months.index(match.group(1)) + 1
  day = int(match.group(2))
  year = int(match.group(3))
  hour = int(match.group(4))
  minute = int(match.group(5))
  half_of_day = match.group(6)
  if half_of_day == 'p.m.':
    hour = (hour + 12) % 24

  return datetime.datetime(year, month, day, hour, minute, 0)


def _parse_release_block(release_block: bs4.Tag) -> model.Release:
  series_link = (release_block
    .find('div', {'class': 'release_title'})
    .find('div', {'class': 'release_text_contents'})
    .find('a'))
  title = series_link.text
  series_id = int(re.match(r'/series/(\d+)/', series_link['href']).group(1))

  subber_link = (release_block
    .find('div', {'class': 'release_subber'})
    .find('div', {'class': 'release_text_contents'})
    .find('a'))
  subber_name = subber_link.text or None
  subber_id = int(re.match(r'/subbertag/(\d+)/', subber_link['href']).group(1)) if subber_name else None

  episode_text = release_block.find('div', {'class': 'release_episode'}).text
  match = re.match(r'(\d+)(?:v(\d+))?', episode_text)
  if match:
    episode = int(match.group(1))
    version = int(match.group(2)) if match.group(2) else None
  else:
    episode = None
    version = None

  quality_prefix = release_block.find(lambda t: t.name == 'b' and t.text == 'Quality:')
  quality = quality_prefix.next_sibling.strip()
  date_string = quality_prefix.parent.find_next_sibling('div', {'class': 'release_last'}).text.strip()
  date = _parse_date_string(date_string)
  extension_prefix = release_block.find(lambda t: t.name == 'b' and t.text == 'Extension:')
  extension = extension_prefix.next_sibling.strip()
  filename = extension_prefix.parent.find_next_sibling('div', {'class': 'release_original'}).text.strip()
  comment_div = release_block.find('div', {'class': 'release_comment_text'})
  comment = comment_div.text.strip() if comment_div else None
  profiles = [x.text for x in release_block.find_all('span', {'class': 'release_profile'})]
  size = release_block.find('div', {'class': 'release_size'}).text
  torrent_link = (
    release_block.find('div', {'class': 'release_download'}) or
    release_block.find('div', {'class': 'release_download_mini'})
  ).parent['href']

  return model.Release(
    int(release_block['id'][3:]),
    model.Subber(subber_id, subber_name) if subber_name else None,
    model.Series(series_id, title),
    episode,
    version,
    model.VideoInfo(extension, filename, quality, profiles, date, comment, size),
    torrent_link)
